/** 
 *   Site testing for SKA candidate.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO:
 *        !) Data save.
 *        *) Agrear configuracion de antenas al filename.
 *
 *
 **/

#define _GNU_SOURCE

//#define FITS_COMPAT
#define NEW_FILE_FORMAT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/file.h>
#include <limits.h>

#include "observ_table.h"
#include "observation.h"
#include "analyzer.h"
#include "my_stdio.h"
#include "my_math.h"
#include "switch_rf.h"
#include "rotator.h"
#include "toska.h"
#include "plot.h"
#include "version.h"
#include "thermometer.h"

#define H_PREFIX "# "

static int az_dif = 0; 

static int mkpath(const char *pathname, mode_t mode)
{
	char *dir, *dirp, *d;
	char *path;
	int path_len = 0;
	int len = strlen(pathname) + 1;
	struct stat m;

	if ( stat(pathname, &m) == 0 ) {
		return 0;
	}

	if ((dir = (char *) malloc(len * sizeof(char))) == NULL) {
		return -1;
	}
	if ((path = (char *) malloc((len + 1) * sizeof(char))) == NULL) {
		return -1;
	}

	d = dir;
	memcpy(dir, pathname, len);
	dir[len - 1] = 0;
	d = strtok_r(dir,"/", &dirp);
	while (dir != NULL) {
		path_len += sprintf(path+path_len, "%s/", dir);
		if ( stat(path, &m) != 0 ) {
			if (mkdir(path, mode) != 0) {
				SYSLOG(LOG_ERR,
				       "mkpath: can not create directory: %s.",
				       path);
				return -1;
			}
		}
		dir = strtok_r(NULL,"/",&dirp);
	}
	free(dir);
	free(path);
	
	return 0;
}

	

static int print_observ_header(int fd, struct obs_setup *s)
{
	struct tm *t;
	
	t = localtime(&s->start);
	fd_printf(fd, H_PREFIX"VERSION = %s\n", VERSION);
	fd_printf(fd, H_PREFIX"NAXIS   = %d\n", 2);
	fd_printf(fd, H_PREFIX"NAXIS1  = %d\n", s->naxis1);
	fd_printf(fd, H_PREFIX"NAXIS2  = %d\n", 1);
#ifdef FITS_COMPAT
	fd_printf(fd, H_PREFIX"CTYPE1  = 'FREQ'\n");
	fd_printf(fd, H_PREFIX"CDELT1  = %14f\n", 
		  (s->obs->stop_freq - s->obs->start_freq)/
		  ((double) (AN_MAX_POINTS - 1)) * 1e6);
	fd_printf(fd, H_PREFIX"CRVAL1  = %14f\n", 
		  s->obs->start_freq * 1e6);

	fd_printf(fd, H_PREFIX"CTYPE2  = 'db'\n");
	fd_printf(fd, H_PREFIX"CDELT2  = %14f\n", 0.0);
	fd_printf(fd, H_PREFIX"CRVAL2  = %14f\n", 1.0);
#endif
	fd_printf(fd, H_PREFIX"DATE = %04d-%02d-%02d\n",
		  t->tm_year + 1900, t->tm_mon + 1, t->tm_mday);
	fd_printf(fd, H_PREFIX"TIME = %02d:%02d:%02d\n",
		  t->tm_hour, t->tm_min, t->tm_sec);
	fd_printf(fd, H_PREFIX"DATE-UNIX = %ld\n", s->start);
	fd_printf(fd, H_PREFIX"AZIMUTH = %3d\n", s->azimuth); 

	fd_printf(fd, H_PREFIX"MODE = %d\n", s->obs->mode); 
	fd_printf(fd, H_PREFIX"START-FREQ = %8.5f\n", s->obs->start_freq); 
	fd_printf(fd, H_PREFIX"STOP-FREQ  = %8.5f\n", s->obs->stop_freq); 
	fd_printf(fd, H_PREFIX"RBW = %7.3f\n", s->obs->RBW); 
	fd_printf(fd, H_PREFIX"SWEEP = %4d\n", s->obs->sweep_obs); 

	fd_printf(fd, H_PREFIX"REF-LEV = %4.0f\n", s->obs->referance_level);
	if( s->obs->VBW > 0)
		fd_printf(fd, H_PREFIX"VBW = %7.2f\n", s->obs->VBW); 
	else
		fd_printf(fd, H_PREFIX"VBW = AUTO\n");
	fd_printf(fd, H_PREFIX"LOG-SCALE = %4.1f\n", s->obs->disp_escale); 


	fd_printf(fd, H_PREFIX"REPETITION = %2d\n", s->obs->repetitions); 
	fd_printf(fd, H_PREFIX"CUR-REP = %2d\n", s->current_rep); 
	fd_printf(fd, H_PREFIX"PARTITION = %2d\n", s->partition); 
	fd_printf(fd, H_PREFIX"ANTENNA-CONF = %c\n", s->obs->switchs); 
	fd_printf(fd, H_PREFIX"POLARIZATION = %c\n", s->polarization); 
	if (thermometer_flag) {
		if (s->temperature > THERMOMETER_FAIL_TEMP)
			fd_printf(fd, H_PREFIX"TEMPERATURE = %6.2f\n", 
				  s->temperature);
		else 
			fd_printf(fd, H_PREFIX"TEMPERATURE = 'FAIL'\n");
	}
	fd_printf(fd, H_PREFIX"END\n");

	return 0;
}


static int save_observ(struct obs_setup *s)
{
	//char file[_POSIX_PATH_MAX];
#ifdef NEW_FILE_FORMAT
	char dir[_POSIX_PATH_MAX];
#endif
	struct tm *t;
	int fd, i;

/* 

Old:

   s2004112617401030h00959999011000004418A
   |                                      \__________
   |                                                 \
   s 2004 11 26 17 40 1 030 h 00959999 01100000 44 18A
   | |    |  |  |  |  | |   | |        |        |  | +-> Antenna Conf. 
   | |    |  |  |  |  | |   | |        |        |  +---> Current repetition
   | |    |  |  |  |  | |   | |        |        +------> Repetitions
   | |    |  |  |  |  | |   | |        +---------------> Stop freq [Hz]
   | |    |  |  |  |  | |   | +------------------------> Start freq [Hz]
   | |    |  |  |  |  | |   +--------------------------> Polarization
   | |    |  |  |  |  | +------------------------------> Azimuth
   | |    |  |  |  |  +--------------------------------> Mode
   | |    |  |  |  +-----------------------------------> Minutes
   | |    |  |  +--------------------------------------> Hour
   | |    |  +-----------------------------------------> Day
   | |    +--------------------------------------------> Month
   | +-------------------------------------------------> Year
   +---------------------------------------------------> ID 


New:

   Directory:

   ../20041126/

   2004 11 26
   |    |  +---> Day
   |    +------> Month
   +-----------> Year

   1-01/00959999-01100000/030h/

   1 - 01 / 00959999 - 01100000 / 030 h /
   |   |    |          |          |   |
   |   |    |          |          |   +---> Polarization
   |   |    |          |          +-------> Azimuth
   |   |    |          +------------------> Stop freq [Hz]
   |   |    +-----------------------------> Start freq [Hz]
   |   +----------------------------------> Partition
   +--------------------------------------> Mode

   File:
   s11015016004418

   s 1101501600 44 18
   | |          |  |
   | |          |  +-> Current repetition
   | |          +----> Repetitions
   | +---------------> seconds since 00:00:00, Jan 1, 1970
   +-----------------> ID

*/
	t = localtime(&s->start);

#ifdef NEW_FILE_FORMAT

	sprintf(dir,"%s/%04d%02d%02d/%1d-%02d/%08.0f-%08.0f/%03d%c",
		s->data_dir,
		t->tm_year + 1900, t->tm_mon + 1, t->tm_mday,
		s->mode,s->partition,
		(s->obs->start_freq * 1000000.0),
		(s->obs->stop_freq * 1000000.0),
		s->azimuth, s->polarization);
	mkpath(dir, S_IRWXU | S_IRGRP | S_IXGRP);
	sprintf(s->file,"%s/s%010ld%02d%02d", 
		dir, s->start, s->obs->repetitions, s->current_rep);

#else
	sprintf(s->file,
		"%s/s%04d%02d%02d%02d%02d%1d%03d%c%08.0f%08.0f%02d%02d%c",
		s->data_dir, t->tm_year + 1900, t->tm_mon + 1, t->tm_mday,
		t->tm_hour, t->tm_min, s->mode, s->azimuth, s->polarization,
		(s->obs->start_freq * 1000000.0),
		(s->obs->stop_freq * 1000000.0),
		s->obs->repetitions, s->current_rep, s->obs->switchs);
#endif

	fd = open(s->file, 
		  O_WRONLY|O_CREAT|O_TRUNC, 
		  S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (fd < 0) {
		SYSLOG(LOG_ERR,"save_observ: Error opening file: %s", s->file);
		return -1;
	}

	print_observ_header(fd, s);
	/* print data */
	for(i=0; i<s->naxis1; i++)
		fd_printf(fd, "%8.2f\n", *(s->data+i));

	close(fd);

	return 0;
	
}


static int get_observ( int dev, struct obs_setup *s)
{
	int ret;
	int i = 0;
	double temperature = 0.0;
	time(&s->start);
	if (thermometer_flag)
		temperature = thermometer_get(s->fd_ther);
	while (analyzer_get_observ( dev, s) != 0) {
		time(&s->start);
		if (thermometer_flag)
			temperature = thermometer_get(s->fd_ther);
		if ((ret = analyzer_reset(dev)) != 0){
		  if (i < AN_RETRY_TIME_OUT) {
		    sleep(1);
		    i++;
		  }
		  else
		    return ret;
		}
	}
	if (thermometer_flag)
		s->temperature = 
			(temperature + thermometer_get(s->fd_ther))/2.0;
	return 0;
}


static int show_current_file(struct obs_setup *s)
{
	int fd;
	char file[_POSIX_PATH_MAX];
	
	sprintf(file,"%s/%s", s->base_dir, CURRENT_FILE);
	fd =open(file, O_WRONLY|O_CREAT|O_SYNC,
		 S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (s->file[0] == '.') {
		fd_printf(fd, "%s%s\n", getenv("PWD"),s->file+1);
	}else{
		fd_printf(fd, "%s\n", s->file);
	}
	close(fd);

	return 0;
}


/**
 * Check and set rotator position
 *
 * log 0: don't log if fail.
 *     1: log if fail.
 * Return:
 *        -1: current position not equal to setup, try to seting this.
 *         0: position is equal to setup, do nothing.
 *
 **/

static int chk_set_az_pol(struct obs_setup *s, int log)
{
	int az;
	char pol;
	int ret = 0;

	if(! rotator_flag)
	        return 0;

	if ((ret = rotator_get_azimuth(s->fd_rot, &az)) != 0) {
		return ret;
	}
	if ((ret = rotator_get_polarization(s->fd_rot, &pol)) != 0) {
		return ret;
	}

	if (  (abs(s->azimuth - az) + az_dif) > ROTATOR_TOLERANCE ||
	      (s->polarization != pol)) {
		if (log) {
			SYSLOG(LOG_ERR,
			       "chk_set_az_pol: current %d, %c; spected %d, %c.\n",
			       az, pol, s->azimuth, s->polarization);
#ifdef BAD_ROTATOR
			/** 
			    This code is for some errors on encoders
			    implementation; this is the "soft"
			    correction.
			**/
			SYSLOG(LOG_ERR,"reseting rotator.");
			rotator_reset(s->fd_rot);
			az_dif = 0;
			/* for exec rotator_set_azimuth() above */
			az = -1;
			/* for exec rotator_set_polarization() above */
			pol = '?';
#endif
		}
		ret = -1;
		if (s->azimuth != az) 
			rotator_set_azimuth(s->fd_rot, s->azimuth);
		if (s->polarization != pol)
			rotator_set_polarization(s->fd_rot, s->polarization);
	}
#ifdef BAD_ROTATOR
	else {
		if (  abs(s->azimuth - az) != 0 ) {
			az_dif += abs(s->azimuth - az);
			SYSLOG(LOG_ERR,"az_dif: %d.", az_dif );
			if ( az_dif > ROTATOR_TOLERANCE ) {
				SYSLOG(LOG_ERR,"reseting rotator.");
				ret = rotator_reset(s->fd_rot);
				az_dif = 0;
			}
		}
	}
#endif
		
	return ret;
}


int observ_mode(int dev, struct obs_setup *s)
{
	struct modes *obsp = s->obs;
	int i, error;

	int done;
#ifdef UPS_LOAD_CTRL
	int rot_tout = 0;
#endif
	while (obsp != NULL && !restart_flag) {
		if (obsp->mode == s->mode) {
			if (antenna_switchs_set(obsp->switchs) < 0) {
				SYSLOG(LOG_ERR, "Cant set rf switch.");
			}
			for(i = 0; i < obsp->repetitions; i++){
				s->current_rep = i+1;
				s->obs = obsp;
				done = 0;
				do {
					switch (chk_set_az_pol(s, 0)){
					case  0:
					case -1: /* Normal on set new
                                                    position */
						done = 1;
#ifdef UPS_LOAD_CTRL
						rot_tout = 0;
#endif
						break;
					case ROT_ERR_TIME_OUT:
						done = 0;
/* FIXME: if rotator not responding, implement reset here.*/
#ifdef UPS_LOAD_CTRL
						rot_tout++;
#endif
						break;
					case ROT_ERR_CONDITION:
					case ROT_ERR_INVALID_PK:
					default:
						done = 0;
						break;
					}
#ifdef UPS_LOAD_CTRL
					/* For testing... */
					if (rot_tout > ROTATOR_TOUT_SUPPORT) {
						rot_tout = 0;
						SYSLOG(LOG_ERR,
						       "Power off load");
						system(LOAD_OFF_CMD);
						sleep(5);
						SYSLOG(LOG_ERR,
						       "Power on load");
						system(LOAD_ON_CMD);
						sleep(1);
						if (antenna_switchs_set(obsp->switchs) < 0) {
							SYSLOG(LOG_ERR, "Cant set rf switch.");
						}
						analyzer_reset(dev);
					}
#endif
				} while(!done);
				done = 0;
				do {
					error = get_observ( dev, s);
					if (error) {
						return error;
					}
					if (chk_set_az_pol(s, 1) == 0) {
						done = 1;
#ifdef UPS_LOAD_CTRL
						rot_tout = 0;
#endif
					}else{
#ifdef UPS_LOAD_CTRL
						rot_tout++;
#endif

						SYSLOG(LOG_ERR, 
						       "rotator position error.");
					}
#ifdef UPS_LOAD_CTRL
					/* For testing... */
					if (rot_tout > ROTATOR_TOUT_SUPPORT) {
						rot_tout = 0;
						SYSLOG(LOG_ERR,
						       "Power off load");
						system(LOAD_OFF_CMD);
						sleep(5);
						SYSLOG(LOG_ERR,
						       "Power on load");
						system(LOAD_ON_CMD);
						sleep(1);
						if (antenna_switchs_set(obsp->switchs) < 0) {
							SYSLOG(LOG_ERR, "Cant set rf switch.");
						}
						analyzer_reset(dev);
					}
#endif
				} while(!done);
				save_observ(s);
				if ( plot_flag )
					plot_observ(s);
				if ( current_flag )
					show_current_file(s);

				free(s->data);

			}
		}
		obsp = obsp->next;
	}

	return 0;
}


struct sequence *observ_sequences(int an_dev, 
				 struct obs_setup *s, struct sequence *seq, 
				 struct modes *obs)
{
	int i;
	struct sequence *seqp = seq;
	struct sequence *seqr = seq;
	int level = seqp->level;
	
	while (seqp != NULL && seqp->type != SEQ_TYPE_NULL && !restart_flag) {
		if( seqp->level != level) {
			if( seqp->level > level) {
				/*
				PVERB(PINFO, 
				      "# loop start %d\n", seqp->level);
				*/
				for(i=0; i<seqp->loop; i++) {
					seqr = observ_sequences(an_dev, s, 
								seqp, obs);
				}
				seqp = seqr;
				/* 
				PVERB(PINFO, "# loop reasume %d\n",level);
				*/
			}else if( seqp->level < level ){
				/*
				PVERB(PINFO, "# loop end %d\n", level);
				*/
				return seqp->next;
			}
		}else{
			if ( seqp->type == SEQ_TYPE_DATA) {
				s->obs = obs;
				s->mode = seqp->mode;
				s->azimuth = seqp->azimuth;
				s->polarization = seqp->polarization;
				s->partition = seqp->partition;
				/*
				PVERB(PINFO, 
				      "# seq = %d-%02d %3d, %d, %d\n", 
				      seqp->mode, seqp->partition, 
				      seqp->azimuth,
				      seqp->level, seqp->loop);
				*/
				observ_mode(an_dev, s);
			}		
			if(seqp->type == SEQ_TYPE_RESTART && 
			   level == 0 ) {
				SYSLOG(LOG_INFO,"restart sequence file.");
				seqp = seq;
			}
			else seqp = seqp->next;
		}
		
	}
	
	return NULL;
}
